@@ -29,6 +29,7 @@ from qpid.messaging import address, transports
 from qpid.messaging.constants import UNLIMITED, REJECTED, RELEASED
 from qpid.messaging.exceptions import *
 from qpid.messaging.message import get_codec, Disposition, Message
+from qpid.messaging.endpoints import MangledString
 from qpid.ops import *
 from qpid.selector import Selector
 from qpid.util import URL, default,get_client_properties_with_defaults
@@ -362,11 +363,11 @@ class Driver:
         [(u.host, default(u.port, 5672)) for u in urls]
     if self._host >= len(hosts):
       self._host = 0
-    result = hosts[self._host]
+    self._last_host = hosts[self._host]
     if self._host == 0:
       self._attempts += 1
     self._host = self._host + 1
-    return result
+    return self._last_host
 
   def _num_hosts(self):
     return len(self.connection.reconnect_urls) + 1
@@ -401,6 +402,24 @@ class Driver:
   def timing(self):
     return self._timeout
 
+  def _check_retry_ok(self):
+    """We consider a reconnect to have suceeded only when we have received
+    open-ok from the peer.
+
+    If we declared success as soon as the transport connected, then we could get
+    into an infinite heartbeat loop if the remote process is hung and never
+    sends us any data. We would fail the connection after 2 missed heartbeats,
+    reconnect the transport, declare the reconnect ok, then fail again after 2
+    missed heartbeats and so on.
+    """
+    if self._retrying and self.engine._connected: # Means we have received open-ok.
+      if self._reconnect_log:
+        log.warn("reconnect succeeded: %s:%s", *self._last_host)
+      self._next_retry = None
+      self._attempts = 0
+      self._delay = self.connection.reconnect_interval_min
+      self._retrying = False
+
   @synchronized
   def readable(self):
     try:
@@ -410,6 +429,7 @@ class Driver:
       elif data:
         rawlog.debug("READ[%s]: %r", self.log_id, data)
         self.engine.write(data)
+        self._check_retry_ok()
       else:
         self.close_engine()
     except socket.error, e:
@@ -451,13 +471,14 @@ class Driver:
     self.schedule()
 
   def update_status(self):
+    if not self.engine: return False
     status = self.engine.status()
     return getattr(self, "st_%s" % status.lower())()
 
   def st_closed(self):
     # XXX: this log statement seems to sometimes hit when the socket is not connected
     # XXX: rawlog.debug("CLOSE[%s]: %s", self.log_id, self._socket.getpeername())
-    self._transport.close()
+    if self._transport: self._transport.close()
     self._transport = None
     self.engine = None
     return True
@@ -483,6 +504,7 @@ class Driver:
   @synchronized
   def timeout(self):
     self.dispatch()
+    self.update_status()
     self._notify()
     self.schedule()
 
@@ -531,12 +553,6 @@ class Driver:
         self._transport = trans(self.connection, host, port)
       else:
         raise ConnectError("no such transport: %s" % self.connection.transport)
-      if self._retrying and self._reconnect_log:
-        log.warn("reconnect succeeded: %s:%s", host, port)
-      self._next_retry = None
-      self._attempts = 0
-      self._delay = self.connection.reconnect_interval_min
-      self._retrying = False
       self.schedule()
     except socket.error, e:
       self.close_engine(ConnectError(text=str(e)))
@@ -589,8 +605,10 @@ class Engine:
     self._status = CLOSED
     self._buf = ""
     self._hdr = ""
-    self._last_in = None
-    self._last_out = None
+    # Set _last_in and _last_out here so heartbeats will be timed from the
+    # beginning of connection if no data is sent/received.
+    self._last_in = time.time()
+    self._last_out = time.time()
     self._op_enc = OpEncoder()
     self._seg_enc = SegmentEncoder()
     self._frame_enc = FrameEncoder()
@@ -658,6 +676,15 @@ class Engine:
 
   def close(self, e=None):
     self._reset()
+    # We cannot re-establish transactional sessions, they must be aborted.
+    # We could re-do transactional enqueues, but not dequeues.
+    for ssn in self.connection.sessions.values():
+      if ssn.transactional:
+        if ssn.committing:
+          ssn.error = TransactionUnknown(text="Transaction outcome unknown due to transport failure")
+        else:
+          ssn.error = TransactionAborted(text="Transaction aborted due to transport failure")
+        ssn.closed = True
     if e:
       self.connection.error = e
     self._status = CLOSED
@@ -706,7 +733,11 @@ class Engine:
     try:
       mech, initial = self._sasl.start(" ".join(mechs))
     except sasl.SASLError, e:
-      raise AuthenticationFailure(text=str(e))
+      if "ANONYMOUS" not in mechs and self.connection.username is None:
+        _text="Anonymous connections disabled, missing credentials"
+      else:
+        _text=str(e)
+      raise AuthenticationFailure(text=_text)
 
     client_properties = get_client_properties_with_defaults(provided_client_properties=self.connection.client_properties);
     self.write_op(ConnectionStartOk(client_properties=client_properties,
@@ -813,13 +844,15 @@ class Engine:
         self.attach(ssn)
         self.process(ssn)
 
-      if self.connection.heartbeat and self._status != CLOSED:
-        now = time.time()
-        if self._last_in is not None and \
-              now - self._last_in > 2*self.connection.heartbeat:
-          raise HeartbeatTimeout(text="heartbeat timeout")
-        if self._last_out is None or now - self._last_out >= self.connection.heartbeat/2.0:
-          self.write_op(ConnectionHeartbeat())
+    # We need to check heartbeat even if not self._connected since we may have
+    # heartbeat timeout before receiving an open-ok
+    if self.connection.heartbeat and self._status != CLOSED and not self._closing:
+      now = time.time()
+      if now - self._last_in > 2*self.connection.heartbeat:
+        raise HeartbeatTimeout(text="heartbeat timeout")
+      # Only send heartbeats if we are connected.
+      if self._connected and now - self._last_out >= self.connection.heartbeat/2.0:
+        self.write_op(ConnectionHeartbeat())
 
   def open(self):
     self._reset()
@@ -906,7 +939,8 @@ class Engine:
           lnk.closed = True
         if _lnk.options.get("delete") in ("always", dir.DIR_NAME):
           dir.do_unlink(sst, lnk, _lnk)
-          self.delete(sst, _lnk.name, unlinked)
+          requested_type = _lnk.options.get("node", {}).get("type")
+          self.delete(sst, _lnk.name, unlinked, node_type=requested_type)
         else:
           dir.do_unlink(sst, lnk, _lnk, unlinked)
         _lnk.closing = True
@@ -922,6 +956,17 @@ class Engine:
         # XXX: subject
         if lnk.options is None:
           lnk.options = {}
+        if isinstance(addr, MangledString):
+          lnk.options['create'] = "always"
+          if 'node' not in lnk.options:
+            lnk.options['node'] = {}
+          if 'x-declare' not in lnk.options['node']:
+            lnk.options['node']['x-declare'] = {}
+          xdeclare = lnk.options['node']['x-declare']
+          if 'auto-delete' not in xdeclare:
+            xdeclare['auto-delete'] = "True"
+          if 'exclusive' not in xdeclare:
+            xdeclare['exclusive'] = "True"
       except address.LexError, e:
         return MalformedAddress(text=str(e))
       except address.ParseError, e:
@@ -940,7 +985,7 @@ class Engine:
       err = None
       if type is None:
         if declare:
-          err = self.declare(sst, lnk, action)
+          err = self.declare(sst, lnk, action, True)
         else:
           err = NotFound(text="no such %s: %s" % (requested_type or "queue", lnk.name))
       else:
@@ -948,9 +993,11 @@ class Engine:
           expected = lnk.options.get("node", {}).get("type")
           if expected and type != expected:
             if declare:
-              err = self.declare(sst, lnk, action)
+              err = self.declare(sst, lnk, action, True)
             else:
               err = AssertionFailed(text="expected %s, got %s" % (expected, type))
+        if "node" in lnk.options and "x-bindings" in lnk.options["node"]:
+          err = self.declare(sst, lnk, action, False)
         if err is None:
           action(type, subtype)
 
@@ -988,6 +1035,11 @@ class Engine:
           type, subtype = None, None
         else:
           type, subtype = "queue", None
+      elif not er.not_found:
+        if node_type == "queue" and force:
+          type, subtype = None, None
+        else:
+          type, subtype = "topic", er.type
       else:
         type, subtype = "topic", er.type
       if type is not None:
@@ -996,41 +1048,45 @@ class Engine:
     sst.write_query(ExchangeQuery(name), do_result)
     sst.write_query(QueueQuery(name), do_action)
 
-  def declare(self, sst, lnk, action):
+  def declare(self, sst, lnk, action, create_node):
     name = lnk.name
     props = lnk.options.get("node", {})
     durable = props.get("durable", DURABLE_DEFAULT)
     type = props.get("type", "queue")
     declare = props.get("x-declare", {})
 
+    cmd = None
     if type == "topic":
-      cmd = ExchangeDeclare(exchange=name, durable=durable)
+      if create_node: cmd = ExchangeDeclare(exchange=name, durable=durable)
       bindings = get_bindings(props, exchange=name)
     elif type == "queue":
-      cmd = QueueDeclare(queue=name, durable=durable)
+      if create_node: cmd = QueueDeclare(queue=name, durable=durable)
       bindings = get_bindings(props, queue=name)
     else:
       raise ValueError(type)
 
-    sst.apply_overrides(cmd, declare)
-
-    if type == "topic":
-      if cmd.type is None:
-        cmd.type = "topic"
-      subtype = cmd.type
+    if cmd is not None:
+      sst.apply_overrides(cmd, declare)
+      if type == "topic":
+        if cmd.type is None:
+          cmd.type = "topic"
+        subtype = cmd.type
+      else:
+        subtype = None
+      cmds = [cmd]
     else:
-      subtype = None
+      cmds = []
 
-    cmds = [cmd]
     cmds.extend(bindings)
 
     def declared():
-      self.address_cache[name] = (type, subtype)
-      action(type, subtype)
+      if create_node:
+        self.address_cache[name] = (type, subtype)
+        action(type, subtype)
 
     sst.write_cmds(cmds, declared)
 
-  def delete(self, sst, name, action):
+  def delete(self, sst, name, action, node_type=None):
     def deleted():
       del self.address_cache[name]
       action()
@@ -1044,7 +1100,7 @@ class Engine:
         action()
       else:
         raise ValueError(type)
-    self.resolve(sst, name, do_delete, force=True)
+    self.resolve(sst, name, do_delete, force=True, node_type=node_type)
 
   def process(self, ssn):
     if ssn.closed or ssn.closing: return
@@ -1312,7 +1368,8 @@ class Engine:
       assert rcv.received < rcv.impending, "%s, %s" % (rcv.received, rcv.impending)
     rcv.received += 1
     log.debug("RCVD[%s]: %s", ssn.log_id, msg)
-    ssn.incoming.append(msg)
+    ssn.message_received(msg)
+
 
   def _decode(self, xfr):
     dp = EMPTY_DP
