@@ -129,52 +129,63 @@ class Connection(Endpoint):
     @rtype: Connection
     @return: a disconnected Connection
     """
-    if url is None:
-      url = options.get("host")
+    # List of all attributes
+    opt_keys = ['host', 'transport', 'port', 'heartbeat', 'username', 'password', 'sasl_mechanisms', 'sasl_service', 'sasl_min_ssf', 'sasl_max_ssf', 'reconnect', 'reconnect_timeout', 'reconnect_interval', 'reconnect_interval_min', 'reconnect_interval_max', 'reconnect_limit', 'reconnect_urls', 'reconnect_log', 'address_ttl', 'tcp_nodelay', 'ssl_keyfile', 'ssl_certfile', 'ssl_trustfile', 'ssl_skip_hostname_check', 'client_properties', 'protocol' ]
+    # Create all attributes on self and set to None.
+    for key in opt_keys:
+        setattr(self, key, None)
+    # Get values from options, check for invalid options
+    for (key, value) in options.iteritems():
+        if key in opt_keys:
+            setattr(self, key, value)
+        else:
+            raise ConnectionError("Unknown connection option %s with value %s" %(key, value))
+
+    # Now handle items that need special treatment or have speical defaults:
+    if self.host:
+        url = default(url, self.host)
     if isinstance(url, basestring):
-      url = URL(url)
+        url = URL(url)
     self.host = url.host
-    if options.has_key("transport"):
-      self.transport = options.get("transport")
-    elif url.scheme == url.AMQP:
-      self.transport = "tcp"
-    elif url.scheme == url.AMQPS:
-      self.transport = "ssl"
-    else:
-      self.transport = "tcp"
+
+    if self.transport is None:
+      if url.scheme == url.AMQP:
+        self.transport = "tcp"
+      elif url.scheme == url.AMQPS:
+        self.transport = "ssl"
+      else:
+        self.transport = "tcp"
     if self.transport in ("ssl", "tcp+tls"):
-      self.port = default(url.port, options.get("port", AMQPS_PORT))
+      self.port = default(url.port, default(self.port, AMQPS_PORT))
     else:
-      self.port = default(url.port, options.get("port", AMQP_PORT))
-    self.heartbeat = options.get("heartbeat")
-    self.username = default(url.user, options.get("username", None))
-    self.password = default(url.password, options.get("password", None))
+      self.port = default(url.port, default(self.port, AMQP_PORT))
+
+    if self.protocol and self.protocol != "amqp0-10":
+        raise ConnectionError("Connection option 'protocol' value '" + value + "' unsupported (must be amqp0-10)")
+      
+    self.username = default(url.user, self.username)
+    self.password = default(url.password, self.password)
     self.auth_username = None
+    self.sasl_service = default(self.sasl_service, "qpidd")
 
-    self.sasl_mechanisms = options.get("sasl_mechanisms")
-    self.sasl_service = options.get("sasl_service", "qpidd")
-    self.sasl_min_ssf = options.get("sasl_min_ssf")
-    self.sasl_max_ssf = options.get("sasl_max_ssf")
+    self.reconnect = default(self.reconnect, False)
+    self.reconnect_interval_min = default(self.reconnect_interval_min,
+                                          default(self.reconnect_interval, 1))
+    self.reconnect_interval_max = default(self.reconnect_interval_max,
+                                          default(self.reconnect_interval, 2*60))
+    self.reconnect_urls = default(self.reconnect_urls, [])
+    self.reconnect_log = default(self.reconnect_log, True)
 
-    self.reconnect = options.get("reconnect", False)
-    self.reconnect_timeout = options.get("reconnect_timeout")
-    reconnect_interval = options.get("reconnect_interval")
-    self.reconnect_interval_min = options.get("reconnect_interval_min",
-                                              default(reconnect_interval, 1))
-    self.reconnect_interval_max = options.get("reconnect_interval_max",
-                                              default(reconnect_interval, 2*60))
-    self.reconnect_limit = options.get("reconnect_limit")
-    self.reconnect_urls = options.get("reconnect_urls", [])
-    self.reconnect_log = options.get("reconnect_log", True)
+    self.address_ttl = default(self.address_ttl, 60)
+    self.tcp_nodelay = default(self.tcp_nodelay, False)
 
-    self.address_ttl = options.get("address_ttl", 60)
-    self.tcp_nodelay = options.get("tcp_nodelay", False)
-
-    self.ssl_keyfile = options.get("ssl_keyfile", None)
-    self.ssl_certfile = options.get("ssl_certfile", None)
-    self.ssl_trustfile = options.get("ssl_trustfile", None)
-    self.ssl_skip_hostname_check = options.get("ssl_skip_hostname_check", False)
-    self.client_properties = options.get("client_properties", {})
+    self.ssl_keyfile = default(self.ssl_keyfile, None)
+    self.ssl_certfile = default(self.ssl_certfile, None)
+    self.ssl_trustfile = default(self.ssl_trustfile, None)
+    # if ssl_skip_hostname_check was not explicitly set, this will be None
+    self._ssl_skip_hostname_check_actual = options.get("ssl_skip_hostname_check")
+    self.ssl_skip_hostname_check = default(self.ssl_skip_hostname_check, False)
+    self.client_properties = default(self.client_properties, {})
 
     self.options = options
 
@@ -209,6 +220,10 @@ class Connection(Endpoint):
           raised this time but won't block future calls
           """
           self.error = None
+          raise e
+
+      if not self._unlinked():
+          self.close()
       raise e
 
   def get_error(self):
@@ -554,6 +569,7 @@ class Session(Endpoint):
     self.closed = False
 
     self._lock = connection._lock
+    self._msg_received = None
 
   def __repr__(self):
     return "<Session %s>" % self.name
@@ -585,6 +601,11 @@ class Session(Endpoint):
     if self.closed:
       raise SessionClosed()
 
+  def message_received(self, msg):
+      self.incoming.append(msg)
+      if self._msg_received:
+          self._msg_received()
+
   @synchronized
   def sender(self, target, **options):
     """
@@ -669,6 +690,18 @@ class Session(Endpoint):
         return msg
     return None
 
+  @synchronized
+  def set_message_received_handler(self, handler):
+      """Register a callback that will be invoked when a message arrives on the
+      session.  Use with caution: since this callback is invoked in the context
+      of the driver thread, it is not safe to call any of the public messaging
+      APIs from within this callback.  The intent of the handler is to provide
+      an efficient way to notify the application that a message has arrived.
+      This can be useful for those applications that need to schedule a task
+      to poll for received messages without blocking in the messaging API.
+      """
+      self._msg_received = handler
+
   @synchronized
   def next_receiver(self, timeout=None):
     if self._ecwait(lambda: self.incoming, timeout):
@@ -717,13 +750,20 @@ class Session(Endpoint):
       raise NontransactionalSession()
     self.committing = True
     self._wakeup()
-    self._ecwait(lambda: not self.committing, timeout=timeout)
+    try:
+      if not self._ecwait(lambda: not self.committing, timeout=timeout):
+        raise Timeout("commit timed out")
+    except TransactionError:
+      raise
+    except Exception, e:
+      self.error = TransactionAborted(text="Transaction aborted: %s"%e)
+      raise self.error
     if self.aborted:
       raise TransactionAborted()
     assert self.committed
 
   @synchronized
-  def rollback(self):
+  def rollback(self, timeout=None):
     """
     Rollback outstanding transactional work. This consists of all
     message sends and receives since the prior commit or rollback.
@@ -732,7 +772,8 @@ class Session(Endpoint):
       raise NontransactionalSession()
     self.aborting = True
     self._wakeup()
-    self._ecwait(lambda: not self.aborting)
+    if not self._ecwait(lambda: not self.aborting, timeout=timeout):
+      raise Timeout("rollback timed out")
     assert self.aborted
 
   @synchronized
@@ -766,9 +807,11 @@ class Session(Endpoint):
     finally:
       self.connection._remove_session(self)
 
+class MangledString(str): pass
+
 def _mangle(addr):
   if addr and addr.startswith("#"):
-    return str(uuid4()) + addr
+    return MangledString(str(uuid4()) + addr)
   else:
     return addr
 
